import { svelte } from '@sveltejs/vite-plugin-svelte';
import fs from 'fs-extra';
import path from 'path';
import url from 'url';
import { generateComponentId, decodeComponentId } from '@previewjs/api';
import { extractCsf3Stories } from '@previewjs/storybook-helpers';
import { maybeOptionalType, intersectionType, objectType, STRING_TYPE, UNKNOWN_TYPE } from '@previewjs/type-analyzer';
import ts from 'typescript';
import { ReaderListeners } from '@previewjs/vfs';
import { preprocess } from 'svelte/compiler';

function analyzeSvelteComponentFromSFC(resolver, virtualSvelteTsAbsoluteFilePath) {
  const sourceFile = resolver.sourceFile(virtualSvelteTsAbsoluteFilePath);
  const propsTypeFields = {};
  let collected = {};
  let slots = [];
  for (const statement of sourceFile?.statements || []) {
    if (ts.isVariableStatement(statement) && statement.modifiers?.find(
      (m) => m.kind === ts.SyntaxKind.ExportKeyword
    ) && statement.declarationList.flags ^ ts.NodeFlags.Const) {
      for (const declaration of statement.declarationList.declarations) {
        if (!ts.isIdentifier(declaration.name)) {
          continue;
        }
        const { type: fieldType, collected: fieldCollected } = resolver.resolveType(
          resolver.checker.getTypeAtLocation(declaration.name)
        );
        propsTypeFields[declaration.name.text] = maybeOptionalType(
          fieldType,
          !!declaration.initializer
        );
        collected = { ...collected, ...fieldCollected };
      }
    } else if (ts.isTypeAliasDeclaration(statement) && statement.name.text === "PJS_Slots") {
      const slotsType = resolver.checker.getTypeAtLocation(statement);
      const resolvedSlotsTypeName = resolver.resolveType(slotsType);
      if (resolvedSlotsTypeName.type.kind === "name") {
        const resolvedSlotsType = resolvedSlotsTypeName.collected[resolvedSlotsTypeName.type.name];
        if (resolvedSlotsType?.type.kind === "tuple") {
          for (const item of resolvedSlotsType.type.items) {
            if (item.kind === "literal" && typeof item.value === "string") {
              slots.push(item.value);
            }
          }
        }
      }
    }
  }
  return {
    propsType: intersectionType([
      objectType(propsTypeFields),
      objectType(
        Object.fromEntries(
          slots.map((slotName) => [`slot:${slotName}`, STRING_TYPE])
        )
      )
    ]),
    types: collected
  };
}

function inferComponentNameFromSveltePath(filePath) {
  return path.basename(filePath, path.extname(filePath));
}

async function extractSvelteComponents(reader, resolver, rootDirPath, absoluteFilePath) {
  if (absoluteFilePath.endsWith(".svelte")) {
    const entry = await reader.read(absoluteFilePath);
    if (entry?.kind !== "file") {
      return [];
    }
    return [
      {
        componentId: generateComponentId({
          filePath: path.relative(rootDirPath, absoluteFilePath),
          name: inferComponentNameFromSveltePath(absoluteFilePath)
        }),
        offsets: [[0, (await entry.read()).length]],
        info: {
          kind: "component",
          exported: true,
          analyze: async () => analyzeSvelteComponentFromSFC(resolver, absoluteFilePath + ".ts")
        }
      }
    ];
  } else {
    const sourceFile = resolver.sourceFile(absoluteFilePath);
    if (!sourceFile) {
      return [];
    }
    return extractCsf3Stories(
      rootDirPath,
      resolver,
      sourceFile,
      async (componentId) => {
        const { filePath } = decodeComponentId(componentId);
        const component = (await extractSvelteComponents(
          reader,
          resolver,
          rootDirPath,
          path.join(rootDirPath, filePath)
        )).find((c) => c.componentId === componentId);
        if (component?.info.kind !== "component") {
          return {
            propsType: UNKNOWN_TYPE,
            types: {}
          };
        }
        return component.info.analyze();
      }
    ).map((c) => {
      if (c.info.kind !== "story" || !c.info.associatedComponent.componentId.includes(".svelte.ts:")) {
        return c;
      }
      const { filePath: associatedComponentFilePath } = decodeComponentId(
        c.info.associatedComponent.componentId
      );
      const associatedComponentSvelteFilePath = stripTsExtension(
        associatedComponentFilePath
      );
      return {
        ...c,
        info: {
          ...c.info,
          associatedComponent: {
            ...c.info.associatedComponent,
            componentId: generateComponentId({
              filePath: associatedComponentSvelteFilePath,
              name: inferComponentNameFromSveltePath(
                associatedComponentSvelteFilePath
              )
            })
          }
        }
      };
    });
  }
}
function stripTsExtension(filePath) {
  return filePath.substring(0, filePath.length - 3);
}

function createSvelteTypeScriptReader(reader) {
  return new SvelteTypeScriptReader(reader);
}
class SvelteTypeScriptReader {
  constructor(reader) {
    this.reader = reader;
    this.listeners = new ReaderListeners();
    reader.listeners.add({
      onChange: (absoluteFilePath, info) => {
        this.listeners.notify(absoluteFilePath, info);
      }
    });
    this.observe = reader.observe?.bind(reader);
  }
  async read(filePath) {
    if (filePath.endsWith(".svelte.ts")) {
      const source = await this.reader.read(
        filePath.substr(0, filePath.length - 3)
      );
      if (source?.kind !== "file") {
        return null;
      }
      return {
        kind: "file",
        name: path.basename(filePath),
        realPath: async () => null,
        lastModifiedMillis: () => source.lastModifiedMillis(),
        read: async () => convertToTypeScript(await source.read()),
        size: () => source.size()
      };
    }
    const entry = await this.reader.read(filePath);
    if (entry?.kind === "directory") {
      return {
        ...entry,
        entries: async () => {
          const realEntries = await entry.entries();
          return Promise.all([
            ...realEntries.map((e) => this.read(path.join(filePath, e.name))),
            ...realEntries.filter((e) => e.name.endsWith(".svelte")).map((e) => this.read(path.join(filePath, e.name + ".ts")))
          ]).then((entries) => entries.filter(Boolean));
        }
      };
    }
    return entry;
  }
  readSync(filePath) {
    if (filePath.endsWith(".svelte.ts")) {
      const source = this.reader.readSync(
        filePath.substr(0, filePath.length - 3)
      );
      if (source?.kind !== "file") {
        return null;
      }
      return {
        kind: "file",
        name: path.basename(filePath),
        realPath: () => null,
        lastModifiedMillis: () => source.lastModifiedMillis(),
        read: () => convertToTypeScript(source.read()),
        size: () => source.size()
      };
    }
    const entry = this.reader.readSync(filePath);
    if (entry?.kind === "directory") {
      return {
        ...entry,
        entries: () => {
          const realEntries = entry.entries();
          return [
            ...realEntries.map(
              (e) => this.readSync(path.join(filePath, e.name))
            ),
            ...realEntries.filter((e) => e.name.endsWith(".svelte")).map((e) => this.readSync(path.join(filePath, e.name + ".ts")))
          ].filter(Boolean);
        }
      };
    }
    return entry;
  }
}
function convertToTypeScript(svelteTemplateSource) {
  let scriptContent = "";
  preprocess(svelteTemplateSource, [
    {
      script: ({ content }) => {
        scriptContent = content;
      }
    }
  ]);
  let slotNames = [];
  let slotSearchPosition = -1;
  while (true) {
    slotSearchPosition = svelteTemplateSource.indexOf(
      "<slot",
      slotSearchPosition + 1
    );
    if (slotSearchPosition === -1) {
      break;
    }
    let slotEndPosition = svelteTemplateSource.indexOf(
      "</slot>",
      slotSearchPosition
    );
    if (slotEndPosition === -1) {
      slotEndPosition = svelteTemplateSource.indexOf("/>", slotSearchPosition);
    }
    const slotCode = svelteTemplateSource.substring(
      slotSearchPosition,
      slotEndPosition
    );
    const slotNameMatch = slotCode.match(/\sname="(.*)"/);
    if (slotNameMatch) {
      slotNames.push(slotNameMatch[1]);
    } else {
      slotNames.push("default");
    }
  }
  const pjsSlotsTypeDeclaration = `type PJS_Slots = [${slotNames.map((slotName) => JSON.stringify(slotName)).join(", ")}];`;
  return `${scriptContent}
${pjsSlotsTypeDeclaration}`;
}

const svelteFrameworkPlugin = {
  isCompatible: async (dependencies) => {
    const version = await dependencies["svelte"]?.readInstalledVersion();
    if (!version) {
      return false;
    }
    return parseInt(version) === 3;
  },
  async create({ rootDirPath }) {
    const __dirname = url.fileURLToPath(new URL(".", import.meta.url));
    const previewDirPath = path.resolve(__dirname, "..", "preview");
    const svelteConfigPath = path.join(rootDirPath, "svelte.config.js");
    let alias = {};
    let isSvelteKit = false;
    if (await fs.pathExists(svelteConfigPath)) {
      const config = (await import(`${url.pathToFileURL(svelteConfigPath).href}?ts=${Date.now()}`)).default;
      alias = config.kit?.alias || {};
      isSvelteKit = Boolean(config.kit);
    }
    return {
      pluginApiVersion: 3,
      name: "@previewjs/plugin-svelte",
      transformReader: (reader) => createSvelteTypeScriptReader(reader),
      defaultWrapperPath: "__previewjs__/Wrapper.svelte",
      previewDirPath,
      detectComponents: async (reader, typeAnalyzer, absoluteFilePaths) => {
        const resolver = typeAnalyzer.analyze(
          absoluteFilePaths.map((p) => p.endsWith(".svelte") ? `${p}.ts` : p)
        );
        const components = [];
        for (const absoluteFilePath of absoluteFilePaths) {
          components.push(
            ...await extractSvelteComponents(
              reader,
              resolver,
              rootDirPath,
              absoluteFilePath
            )
          );
          await 0;
        }
        return components;
      },
      viteConfig: (configuredPlugins) => ({
        ...isSvelteKit ? {
          define: {
            __SVELTEKIT_DEV__: "false",
            __SVELTEKIT_APP_VERSION_POLL_INTERVAL__: "0"
          },
          resolve: {
            alias: {
              $app: "node_modules/@sveltejs/kit/src/runtime/app",
              ...alias
            }
          },
          publicDir: "static"
        } : {},
        plugins: [
          ...configuredPlugins.filter(
            (plugin) => plugin.name !== "vite-plugin-svelte-kit" && plugin.name !== "vite-plugin-sveltekit-build" && plugin.name !== "vite-plugin-sveltekit-middleware" && plugin.name !== "vite-plugin-sveltekit-setup" && plugin.name !== "vite-plugin-sveltekit-compile"
          ),
          configuredPlugins.find((plugin) => plugin.name.includes("svelte")) ? null : svelte(),
          isSvelteKit ? {
            name: "previewjs:fake-sveltekit-client",
            transform(code, id) {
              if (id.includes(
                "@sveltejs/kit/src/runtime/client/singletons.js"
              )) {
                return code.replace(`export let client`, `export let client = {}`).replace(
                  `url: notifiable_store({})`,
                  `url: notifiable_store(document.location)`
                ).replace(
                  `page: notifiable_store({})`,
                  `page: notifiable_store({ url: document.location })`
                );
              }
              return null;
            }
          } : null,
          {
            name: "previewjs:disable-svelte-hmr",
            async transform(code, id) {
              if (!id.endsWith(".svelte")) {
                return null;
              }
              return code.replace(/import\.meta\.hot/g, "false");
            }
          }
        ]
      })
    };
  }
};

export { svelteFrameworkPlugin as default };
